/* -*- Mode: javascript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-breakpoints.js -- manage breakpoints
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

// This class is a singleton
JSDBGP.BreakpointManager = function(mainObj) {
    // use an array to map breakpoint IDs to breakpoints
    this.breakpoints = [];
    this._nextFreeBreakpointID = 1; // Skip 0
    
    this.mainObj = mainObj;

    this.log = JSDBGP.Logging.getLogger("winger.breakpointManager");
    this.log.setLevel(JSDBGP.Logging.DEBUG);
};

JSDBGP.BreakpointManager.prototype.finalInit = function() {
    this.bpsMediator = this.mainObj.bpsMediator;
};

// Input: args object from command
// Output: breakpoint ID, or throw an exception
JSDBGP.BreakpointManager.prototype.addFileBreakpoint = function(args_obj) {
    if (!args_obj.filename) {
        this.mainObj.console.command.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
        throw new Error("BreakpointManager.addFileBreakpoint : no file specified");
    }
    if (!args_obj.lineno) {
        this.mainObj.console.command.error_code = JSDBGP.Constants.E_INVALID_OPTIONS;
        throw new Error("BreakpointManager.addFileBreakpoint : no lineno specified");
    }

    var urlType = this.mainObj.utils.classifyURL(args_obj.filename);
    var mapMode_orig, mapMode_other;
    var tagTest;
    var bi_pp, bi_src, pc, script;
    var is_pp_uri;
    switch (urlType) {
    case JSDBGP.Constants.JavaScriptURL:
    case JSDBGP.Constants.JavaScriptWithBase:
    case JSDBGP.Constants.JavaScriptIndex:
        is_pp_uri = true;
        break;
    default:
        is_pp_uri = false;
    }

    if (is_pp_uri) {
        var retObj = {};
        if (!this.mainObj.console.command.extractInfoFrom_DBGP_URI(args_obj.filename, retObj)) {
            this.mainObj.console.command.error_code = JSDBGP.Constants.E_INTERNAL_ERROR;
            throw new Error("Can't get tag out of url " + url);
        }
            
        var tag = retObj.tag;
        if (tag <= 0 || !(tag in this.mainObj.console.scriptWrappers)) {
            this.mainObj.console.command.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID;
            throw new Error("BreakpointManager.addFileBreakpoint : script with tag [" + tag + "] no longer exists");
        }
        bi_pp = this._createBreakpointAux(args_obj.filename, args_obj.lineno,
                                      JSDBGP.Constants.PCMAP_PRETTYPRINT,
                                      args_obj,
                                      false);
        if ((pc = this.bind_DBGP_Breakpoint(bi_pp, tag)) != null) {
            if (urlType != JSDBGP.Constants.JavaScriptURL) {
                var sw = this.mainObj.console.scriptWrappers[tag];
                var jsdScript = sw ? sw.jsdScript : null;
                if (jsdScript) {
                    var mapMode = JSDBGP.Constants.PCMAP_SOURCETEXT;
                    var srcLine = jsdScript.pcToLine(pc, mapMode);
                    bi_src = this._createBreakpointAux(jsdScript.fileName,
                                                       srcLine,
                                                       mapMode,
                                                       args_obj,
                                                       true);
                    bi_src.other = bi_pp;
                    bi_pp.other = bi_src;
                    var command = this.mainObj.console.command;
                    command.addNotification(function() {
                        command.notifyBreakpointSet(bi_src.bkpt_id);
                    }, false);
                }
            }
        }
        return bi_pp.bkpt_id;
    } else {
        var old_id_hash = this.bpsMediator.getSourceBreakpointAtLine(args_obj.filename, args_obj.lineno);
        for (var old_id in old_id_hash) {
            if (old_id) {
                this.removeBreakpoint(old_id);
            }
        }
        bi_src = this._createBreakpointAux(args_obj.filename, args_obj.lineno,
                                           JSDBGP.Constants.PCMAP_SOURCETEXT,
                                           args_obj, true);
        this.bindSrcBreakpointWrapper(bi_src);
        return bi_src.bkpt_id;
    }
};

// Precondition: we just created a pretty-print source and want to
// copy a set of breakpoints.  This is the only point where this
// happens. 

JSDBGP.BreakpointManager.prototype.cloneBreakpoints = function(tag, jsdScript) {



    var pc_id_list = this.bpsMediator.getBoundSrcBkpts(tag);
    for (var pc in pc_id_list) {
        var bkpt_id = pc_id_list[pc];
        if (bkpt_id) {
            this.cloneBreakpoint(tag, jsdScript, bkpt_id);
        }
    }
}

JSDBGP.BreakpointManager.prototype.cloneBreakpoint = function(tag, jsdScript, bkpt_id) {
    var bi = this.getBreakpoint(bkpt_id);
    if (!bi) {



        return;
    }
    if (!(tag in this.mainObj.console.pushedSources)) {



        return;
    }
    var new_url = this.mainObj.console.pushedSources[tag];
    var mapMode = JSDBGP.Constants.PCMAP_PRETTYPRINT;
    var pc = this.bpsMediator.getPC(tag, bi.bkpt_id);
    if (pc == null) {



        return;
    }
    var lineno = jsdScript.pcToLine(pc, mapMode);
    var bkpt_new_id = this._nextFreeBreakpointID++;
    



    var bi_new = new JSDBGP.Breakpoint(new_url, lineno, mapMode, null, bkpt_new_id, this.mainObj);
    this.breakpoints[bkpt_new_id] = bi_new;
    this.bpsMediator.reportBreakpointAdd(new_url, lineno, bi_new.bkpt_id);
    var props = ['state', 'stateVal', 'condition', 'hit_count', 'hit_value', 'hit_condition_func', 'hit_condition'];
    for (var idx2 in props) {
        var prop = props[idx2];
        if (prop in bi && !(prop in bi_new)) {
            bi_new[prop] = bi[prop];
        }
    }
    bi.other = bi_new;
    bi_new.other = bi;
    this.bpsMediator.reportAddPPBkpt(tag, bkpt_new_id, pc);
    var command = this.mainObj.console.command;



    command.addNotification(function() {
        command.notifyBreakpointSet(bkpt_new_id);
    }, false);
};

JSDBGP.BreakpointManager.prototype.getAllBreakpoints = function() {
    var ret = [];
    var urls = this.bpsMediator.getSourceActiveURLs(url);
    for (var i in urls) {
        var url = urls[i];
        var lines = this.bpsMediator.getSourceBreakpoints(url);
        for (var line in lines) {
            var id_hash = lines[line];
            for (var bkpt_id in id_hash) {
                var bi = this.breakpoints[bkpt_id];
                if (bi) {
                    ret.push(bi);
                }
            }
        }
    }
    return ret;
};

JSDBGP.BreakpointManager.prototype.getBreakpoint = function(bkpt_id) {
    if (!(bkpt_id in this.breakpoints)) {
        return null;
    }
    return this.breakpoints[bkpt_id];
};

JSDBGP.BreakpointManager.prototype.removeAllBreakpoints = function() {



    for (var bkpt_id in this.breakpoints) {
        this.removeBreakpoint(bkpt_id);
    }
    this.clearBreakpoints();



};

//precondition: neither bi nor bi.other are null
JSDBGP.BreakpointManager.prototype._breakAssociation = function(bi) {
    bi.other.other = null;
    bi.other = null;
};

// Can be called from the mediator
JSDBGP.BreakpointManager.prototype.removeBreakpointByTag = function(tag, bkpt_id) {
    var bi = this.getBreakpoint(bkpt_id);
    if (bi) {
        if (bi.other) {
            this._breakAssociation(bi);
        }
        var sw = this.mainObj.console.scriptWrappers[tag];
        if (sw) {
            var jsdScript = sw.jsdScript;
            if (jsdScript.isValid) {
                var pc = this.bpsMediator.getPC(tag, bkpt_id);
                if (pc != null) {
                    jsdScript.clearBreakpoint(pc);
                }
            }
        }
    }
    delete this.breakpoints[bkpt_id];
};

JSDBGP.BreakpointManager.prototype.removeBreakpoint = function(bkpt_id) {
    var bi = this.getBreakpoint(bkpt_id);
    if (!bi) {



        return;
    }
    var id2 = 0;
    if (bi && bi.other) {
        id2 = bi.other.bkpt_id;
        this._breakAssociation(bi);
    }
    this._removeBreakpointAux(bkpt_id, true);
    if (id2 != 0) {



        this._removeBreakpointAux(id2, false);
        var command = this.mainObj.console.command;
        command.addNotification(function() {
            command.notifyBreakpointDelete(id2);
        }, false);
    }
};

JSDBGP.BreakpointManager.prototype.notifyBreakpointRemove = function(bkpt_id) {
    var command = this.mainObj.console.command;
    command.addNotification(function() {
        command.notifyBreakpointDelete(bkpt_id);
    }, false);
};

JSDBGP.BreakpointManager.prototype._removeBreakpointAux = function(bkpt_id, clear_breakpoint)  {
    var tags = this.bpsMediator.getBoundTags(bkpt_id);
    for (var i in tags) {
        var tag = tags[i];
        if (clear_breakpoint) {
            this.removeBreakpointByTag(tag, bkpt_id);
        } else {
            // high coupling here -- removeBreakpointByTag removes bkpts from
            // the list, but if we don't call it we still need to do it.
            delete this.breakpoints[bkpt_id];
        }
    }
    this.bpsMediator.reportBkptDelete(bkpt_id);
};

JSDBGP.BreakpointManager.prototype.testBreakpoint = function(tag, currentFrame, defaultValue) {
    var pc = currentFrame.pc;
    var bkpt_id = this.bpsMediator.getAnyBkpt(tag, currentFrame.pc);
    if (bkpt_id) {
        var bi = this.getBreakpoint(bkpt_id);
        if (bi) {
            return bi.testBreakpoint(tag, currentFrame, defaultValue);
        }
    }
    return defaultValue;
}

JSDBGP.BreakpointManager.prototype.updateBreakpoint = function(bkpt_id, args_obj) {
    this._updateBreakpointAux(bkpt_id, args_obj);
    var bi = this.getBreakpoint(bkpt_id);
    if (bi) {
        var bi_other = bi.other;
        if (bi_other) {
            var command = this.mainObj.console.command;
            var other_id = bi_other.bkpt_id;
            this._updateBreakpointAux(other_id, args_obj);
            command.addNotification(function() {
                command.notifyBreakpointModify(other_id);
            }, false);
        }
    }
};

JSDBGP.BreakpointManager.prototype._updateBreakpointAux = function(bkpt_id, args_obj) {
    var bi = this.getBreakpoint(bkpt_id);
    if (bi) {
        if (args_obj.state) {
            bi.updateState(args_obj.state);
        }
        bi.updateHitCondition(args_obj);
        if ('lineno' in args_obj) {
            this.mainObj.console.command.error_code = JSDBGP.Constants.E_BREAKPOINT_TYPE_NOT_SUPPORTED;
            throw new Error("Not ready to handle updating other parts of a breakpoint");
        }
        return true;
    } else {
        this.log.debug("Breakpoint " + bkpt_id + " can't be found");
    }
    return false;
};

JSDBGP.BreakpointManager.prototype.bindSrcBreakpointWrapper = function(bi) {
    var console = this.mainObj.console;
    var sm = console.scriptManagers[bi.filename];
    var retVal = false;
    if (!sm) {



        // We're setting the breakpoint before the script manager has
        // been created.  
        // No script to bind against.
        return;
    }
    for (var tag in sm.scriptWrappers) {
        var pc = this.bindSrcBreakpoint(bi, tag);
        var sw = sm.scriptWrappers[tag];
        var jsdScript = null;
        if (sw) {
            jsdScript = sw.jsdScript;
        }
        if (pc && jsdScript && tag in console.pushedSources) {
            this.mainObj.breakpointManager.cloneBreakpoint(tag, jsdScript, bi.bkpt_id);
        }
    }
};

JSDBGP.BreakpointManager.prototype.bindSrcBreakpoint = function(bi, tag) {
    if (this.mainObj.utils.memberOf(tag, this.bpsMediator.getBoundTags(bi.bkpt_id))) {
        // We already have a breakpoint -- this happens on
        // subsequent runs when scripts aren't trashed.





        return null;
    }
    var pcmap_mode = JSDBGP.Constants.PCMAP_SOURCETEXT;
    var console = this.mainObj.console;
    var sm = console.scriptManagers[bi.filename];
    if (!sm) {




    }
    var sw = sm.scriptWrappers[tag];
    if (!sw) {



        return null;
    }
    var jsdScript = sw.jsdScript;
    if (!jsdScript || !jsdScript.isValid) {



        return null;
    }
    try {
        if (jsdScript.baseLineNumber <= bi.lineno
            && bi.lineno <= (jsdScript.baseLineNumber + jsdScript.lineExtent)
            && jsdScript.isLineExecutable(bi.lineno, pcmap_mode)) {
            var pc = jsdScript.lineToPc(bi.lineno, pcmap_mode);
            if (pc < 0) {




                return null;
            }
                
            this.bpsMediator.reportAddSrcBkpt(tag, bi.bkpt_id, pc);
            if (tag in console.pushedSources) {
                this.cloneBreakpoint(tag, jsdScript, bi.bkpt_id);
            }
            if (bi.stateVal == JSDBGP.Constants.BKPT_ENABLE
                || bi.stateVal == JSDBGP.Constants.BKPT_TEMPORARY) {












                jsdScript.setBreakpoint(pc);
            } else {



            }
            return pc;
        }
    } catch(ex) {
        this.log.exception(ex);
    }
    return null;
};

JSDBGP.BreakpointManager.prototype.bind_DBGP_Breakpoint = function(bi, tag) {
    var sw = this.mainObj.console.scriptWrappers[tag];
    var jsdScript = sw ? sw.jsdScript : null;
    if (!jsdScript || !jsdScript.isValid) {



        return null;
    }
    var pcmap_mode = JSDBGP.Constants.PCMAP_PRETTYPRINT;
    try {
        if (jsdScript.isLineExecutable(bi.lineno, pcmap_mode)) {
            var pc = jsdScript.lineToPc(bi.lineno, pcmap_mode);
            this.bpsMediator.reportAddPPBkpt(tag, bi.bkpt_id, pc);
            if (bi.stateVal == JSDBGP.Constants.BKPT_ENABLE) {
                jsdScript.setBreakpoint(pc);
            }
            return pc;
        }
    } catch(ex) {
        this.log.exception(ex);
    }
    return null;
};

JSDBGP.BreakpointManager.prototype._createBreakpointAux = function(url, lineno, mapMode, args_obj, atomic) {
    var bkpt_id, bi;
    if (!atomic) {
        var old_id_hash = this.bpsMediator.getSourceBreakpointAtLine(url, lineno);
        for (bkpt_id in old_id_hash) {
            bi = bkpt_id ? this.breakpoints[bkpt_id] : null;
            if (bi) {
                this.removeBreakpoint(bkpt_id);
            }
        }
    }
    bkpt_id = this._nextFreeBreakpointID++;
    this.bpsMediator.reportBreakpointAdd(url, lineno, bkpt_id);
    bi = new JSDBGP.Breakpoint(url, lineno, mapMode, args_obj, bkpt_id, this.mainObj);
    this.breakpoints[bkpt_id] = bi;
    return bi;
};

JSDBGP.BreakpointManager.prototype.clearBreakpoints = function() {
    // use an array to map breakpoint IDs to breakpoints
    this.breakpoints = [];        
}

// These classes are instantiated multiple times, so it makes sense to use
// prototypes.

// See commands.breakpoint_set for description of args_obj

JSDBGP.Breakpoint = function(url, lineno, mapMode, args_obj, bkpt_id, mainObj) {
    this.bkpt_id = bkpt_id;
    this.filename = url;
    this.lineno = lineno;
    this.mapMode = mapMode;
    this.mainObj = mainObj;
    this.log = this.mainObj.breakpointManager.log;
    if (args_obj) {
        this.state = args_obj.state;
        if (args_obj.is_temporary && this.state == "enabled") {
            this.state = "temporary";
        }
        this.stateVal = this.stateToVal(this.state);
        if ('expression' in args_obj) {
            this.condition = args_obj.expression;
        } else {
            this.condition = null;
        }
        this.updateHitCondition(args_obj);
    } else {
        // We're cloning, and we'll set the other properties in the caller.
    }
};

JSDBGP.Breakpoint.prototype.stateToVal = function(state) {
    switch (state) {
    case 'enabled':
        return JSDBGP.Constants.BKPT_ENABLE;
    case 'disabled':
        return JSDBGP.Constants.BKPT_DISABLE;
    case 'temporary':
        return JSDBGP.Constants.BKPT_TEMPORARY;
    }
    this.mainObj.console.command.error_code = JSDBGP.Constants.E_BREAKPOINT_INVALID_STATE;
    throw new Error("JSDBGP.Breakpoint.prototype.stateToVal: invalid state name of " + (state ? state : "<undefined>"));
}

JSDBGP.Breakpoint.prototype.hit_funcs = {
    '>=' : function(hit_count, hit_value) { return hit_count >= hit_value },
    '==' : function(hit_count, hit_value) { return hit_count == hit_value },
    // precondition: hit_count > 0 on every call due to pre-increment
    '%' : function(hit_count, hit_value) { return hit_count % hit_value == 0 },
    __end__ : null
};

JSDBGP.Breakpoint.prototype.display_e4x = function() {
    var breakpoint = <breakpoint />;
    breakpoint.@type = "line";
    breakpoint.@state = this.state;
    breakpoint.@filename = this.mainObj.scripts.mapHTTP_URI_to_FileURI(this.filename);
    breakpoint.@lineno = this.lineno;
    if (this.hit_value) {
        breakpoint.@hit_value = this.hit_value;
        breakpoint.@hit_condition = this.hit_condition;
        breakpoint.@hit_count = this.hit_count;
    }
    return breakpoint;
}

JSDBGP.Breakpoint.prototype.is_enabled = function() {
    return this.stateVal != JSDBGP.Constants.BKPT_DISABLE;
};

JSDBGP.Breakpoint.prototype.enable_disable = function(whichWay) {
    var fn_name;
    var tag;
    var sw_list = this.sw_list;
    for (tag in sw_list) {
        try {
            var entry = sw_list[tag];
            var sw = entry[0];
            var pc = entry[1];
            var jsdScript = sw.jsdScript;
            if (jsdScript.isValid) {
                if (whichWay) {
                    this.stateVal = JSDBGP.Constants.BKPT_ENABLE;
                    this.state = "enabled";
                    fn_name = "setBreakpoint";
                } else {
                    this.stateVal = JSDBGP.Constants.BKPT_DISABLE;
                    this.state = "disabled";
                    fn_name = "clearBreakpoint";
                }



                (jsdScript[fn_name])(pc);
            }
        } catch(ex) {
            this.log.exception(ex);
        }
    }
    return true;
};

JSDBGP.Breakpoint.prototype.disableAt_PC = function(pc) {
    var tag;
    var sw_list = this.sw_list;
    for (tag in sw_list) {
        try {
            var entry = sw_list[tag];
            if (pc == entry[1]) {
                var jsdScript = entry[0].jsdScript;
                if (jsdScript.isValid) {
                    jsdScript.clearBreakpoint(pc);
                }
            }
        } catch(ex) {
            this.log.exception(ex);
        }
    }
        //@@@@
        // percolate to other bi
};

JSDBGP.Breakpoint.prototype.testBreakpoint = function(currentFrame, pc, defaultValue) {
    // This function returns defaultValue to signal it's time to break,
    // and one of the RETURN_* values if it isn't.
    
    if (this.stateVal == JSDBGP.Constants.BKPT_DISABLE) {
        return JSDBGP.Constants.RETURN_CONTINUE;
    }
    // If it's a temporary breakpoint, chances are good it was
    // created by the run-to-cursor command, and it's a standard
    // file+line bpt with no conditions, hit-tests, etc.

    else if (this.stateVal == JSDBGP.Constants.BKPT_TEMPORARY) {
        this.stateVal = JSDBGP.Constants.BKPT_DISABLE;
        this.disableAt_PC(pc);
        return defaultValue;
    }

    if (this.hit_value) {
        this.hit_count += 1;
        if (this.hit_condition_func(this.hit_count, this.hit_value)) {
            return defaultValue;
        }
        // This condition failed, so don't break.
        return JSDBGP.Constants.RETURN_CONTINUE;
    }

    if (this.condition != null) {
        var result = new Object();
        if (!currentFrame.eval(this.condition,
                               this.mainObj.console.command.dbgp_eval_ns + "/breakpoint-condition",
                               1, result)) {
            // Some kind of failure, so just break here.
            return defaultValue;
        } else if (result.value.getWrappedValue()) {
            return defaultValue;
        } else {
            return JSDBGP.Constants.RETURN_CONTINUE;
        }
    }
    return defaultValue;
};

JSDBGP.Breakpoint.prototype.updateState = function(newState) {
    if (this.state != newState) {
        this.state = newState;
        this.stateVal = this.stateToVal(newState);
        this.enable_disable(this.stateVal == JSDBGP.Constants.BKPT_ENABLE);
    }
};

JSDBGP.Breakpoint.prototype.updateHitCondition = function(args) {
    // Set the hit-count back to 0 if either the target
    // or the condition has changed.
    
    if ('hit_value' in args) {
        this.hit_count = 0;
        this.hit_value = args.hit_value;
        if (args.hit_value == 0) {
            // Setting <hit_value = 0> means remove the hit count condition.
            delete this.hit_condition_func;
            delete this.hit_condition;
        }
    }
    if ('hit_condition' in args) {
        this.hit_count = 0;
        this.hit_condition = args.hit_condition;
        this.hit_condition_func = this.hit_funcs[this.hit_condition];
    }
};
