/* -*- Mode: JavaScript; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * Copyright (c) 2006, 2007 ActiveState Software Inc.
 * Refer to the license in the Komodo IDE installation
 * for the licensing information of this source code.
 *
 * winger-debugger.js -- manage the URL and script objects
 *
 * Ideas borrowed from Venkman, but we aren't using future breakpoints,
 * and we can have more than one script associated with a breakpoint at any
 * time.
 */

if (typeof(JSDBGP) == 'undefined') {
    var JSDBGP = {};
}

/* Classes:
 *
 * + ScriptManager: wrapper around each URL that manages source code and scripts
 *
 * + URLInstance : This is the complex part of Venkman, but handles more
 * info related to a URL (file)
 *
 * + ScriptWrapper : smart wrapper around jsdIScript
 *
 * Objects
 *
 * console -- the main object that wraps the debugger session
 * should be called "session", using "console" because Venkman does.
 */


JSDBGP.Debugger = function(mainObj) {
    this.mainObj = mainObj;
    this.console = mainObj.console;
    this.reached_base_frame = false;
    this.gave_message = {};
    this.XML_saved_settings = {};
    this.log = JSDBGP.Logging.getLogger("winger.debugger");
    this.log.setLevel(JSDBGP.Logging.DEBUG);

    this.MozPref = new Components.Constructor("@mozilla.org/preferences;1","nsIPref");

    this.initDebugger = function (viewObj, host_name, host_port,
                                  session_key, in_interactive_shell) {
        try {
            var console = this.console;
            console.command.in_interactive_shell = in_interactive_shell;
            console.command.session_key = session_key;

            // Clean up from previous runs
            if (console.frames) {
                delete console.frames;
            }
            




            if (console.session) {
                this.log.warn("console.session still alive!  ending previous session");
                console.session.end_session();
            }
            
            console.session = new JSDBGP.Session(console.command, viewObj, host_name, host_port);
            
            // Do inter-object initialization here,
            // like callback registration

            this.mainObj.scripts.finalInit();

            this.pendingScripts = {}; // hash on jsdScript.tag
            this.hooksOffScripts = [];
            
            /* create the debugger instance */

            var utils = this.mainObj.utils;
            if (!(JSDBGP.Constants.JSD_CTRID in Components.classes)) {
                throw new Error("JavaScript debugger (" +  JSDBGP.Constants.JSD_CTRID + ") not available");
            }
            console.jsds = Components.classes[JSDBGP.Constants.JSD_CTRID].getService(Components.interfaces.jsdIDebuggerService);
            console.jsds.on();




            // We have to push the current object into the sub-objects
            // so the intermediate methods can invoke the methods in
            // the main debugger object.

            var self = this;

            console.breakpointHook = {
                onExecute: function(frame, type, rv) {
                    return self.jsdBreakpointHook(frame, type, rv);
                }
            };
            console.executionHook = {
                onExecute : function(frame, type, rv) {
                    return self.jsdExecutionHook(frame, type, rv);
                }
            };
                
            console.executionHook_prologue = {
                onExecute : function(frame, type, rv) {
                    return self.jsdExecutionHook_prologue(frame, type, rv);
                }
            };
            console.errorHook = null;
            /*
            {
                onError: function(message, fileName, line, pos, flags, exception) {
                    return self.jsdErrorHook(message, fileName, line, pos, flags, exception);
                }
            };
            */
            console.callHook = {
                onCall: function(frame, type) {
                    return self.jsdCallHook(frame, type);
                }
            };
            console.scriptHook = {
                scriptCreatedName : "jsdScriptCreated_prologue",
                scriptDestroyedName : "jsdScriptDestroyed_prologue",
                onScriptCreated : function(jsdScript) {
                    // return self[this.scriptCreatedName](jsdScript);
                    var self_c_sh = this;  // for self.console.scriptHook
                    setTimeout(function() {
                        // Make sure the hook wasn't cleared in between
                        // invocation and when the hook triggers.
                        var fn = self[self_c_sh.scriptCreatedName];
                        if (fn) {
                            fn.apply(self, [jsdScript]);
                        }
                    }, 1);
                },
                onScriptDestroyed : function(jsdScript) {
                    // No need to wait here.
                    return self[this.scriptDestroyedName].apply(self, [jsdScript]);
                }
            };

            console.jsds.scriptHook = console.scriptHook;
            var enumer = {
                enumerateScript: function(jsdScript) {











                    self.newScriptsByTag[jsdScript.tag] = jsdScript;
                }
            };
            this.newScriptsByTag = {};
            console.jsds.enumerateScripts(enumer);



            this._updateNewScripts();

            console.jsds.breakpointHook = console.breakpointHook;
            console.jsds.debuggerHook   = console.executionHook;
            console.jsds.debugHook      = console.executionHook;
            console.jsds.errorHook      = console.errorHook;



            this.clearInterruptHook();
            // console.jsds.interruptHook      = console.executionHook_prologue;



            console.jsds.topLevelHook = console.jsds.functionHook = null;
            // Doing this is useless -- topLevelHooks don't give us much info.

            console.jsds.flags = Components.interfaces.jsdIDebuggerService.ENABLE_NATIVE_FRAMES;

            this.saved_hook_stack = [];

            console.walkedFrames = {};
            this.sentInitString_p = false;

            // Check the hooks
            for (var p in console.jsds) {
                var hookLocn = p.search("Hook");
                if (hookLocn >= 0 && hookLocn == p.length - 4) {



                }
            }
            this.setInitialInterrupt = false;

            this._saveSettings();

            {
                var self = this;
                var command = console.command;
                const stop_reason = (in_interactive_shell ?
                                     JSDBGP.Constants.STOP_REASON_INTERACT :
                                     JSDBGP.Constants.STOP_REASON_STARTING);
                setTimeout(function() {
                    command.sendInitString(stop_reason);
                    command.doEnterNestedEventLoop(stop_reason);
                }, 1);
                self.sentInitString_p = true;
            }




        } catch(ex) {
            this.log.exception(ex);
            this.cleanup();
            throw ex;
        }
    };

    this.cleanup = function() {
        // in failure cases, we want to be sure we really cleanup any hooks
        // or objects we've created.  We may want to grab some of the code
        // from detachDebugger below.
        if (console.session) {
            console.session.end_session();
            console.session = null;
        }
    };

    this._clearHooks = function(jsds) {
        jsds.topLevelHook = null;
        jsds.functionHook = null;
        jsds.breakpointHook = null;
        jsds.debuggerHook = null;
        jsds.errorHook = null;
        this.clearInterruptHook();
    };

    this.detachDebugger = function() {
        var console = this.console;
        if (!console) {
            try {
                this.log.debug("detachDebugger: no console...");
            } catch(ex) {
            }
            return;
        }
        try {
            this.mainObj.breakpointManager.removeAllBreakpoints();
            this._clearHooks(console.jsds);

            // Don't turn the scriptHook off!  Once it's on it
            // needs to stay on to track script changes.
            // **DON'T UNCOMMENT** console.jsds.scriptHook = null;

            try {
                console.jsds.clearAllBreakpoints();
            } catch(ex) {
                if (ex.name != "NS_ERROR_NOT_AVAILABLE") {
                    this.log.exception(ex);
                } else {
                    this.log.debug("Can't clear breakpoints");
                }
            }
        } catch(ex) {
            this.log.exception(ex);
        }
        if (!console.initialized) {



            // return;
        }
        try {



            while (console.nested_event_loop_depth > 0) {
                var times = 0;



                console.command.doExitNestedEventLoop();
                if (++times >= 10) {



                    break;
                }
            }
            this.clearStepPast();
            this.mainObj.scripts.finalShutdown();
            JSDBGP.winger_reset_ui(JSDBGP.Constants.DBGP_STATE_DISCONNECTED);
        } catch(ex) {
            this.log.exception(ex);
        }
        if (console.session) {
            console.session.end_session();
            console.session = null;
        }
        this._restoreSettings();
        console.initialized = false;



    };

    this._saveSettings = function() {
        // We don't want the mozilla notice of scripts taking too long
        // while our app is waiting for a continuation command from Komodo
        
        var _mozPref = new this.MozPref();
        this.orig_max_script_run_time = _mozPref.GetIntPref('dom.max_script_run_time');
        _mozPref.SetIntPref('dom.max_script_run_time', 0);
        delete _mozPref;

        // We don't want pretty-printing for toXMLString()
        // because Python treats white-space nodes as significant.
        var ds = XML.settings();
        this.XML_saved_settings.prettyPrinting = ds.prettyPrinting;
        this.XML_saved_settings.prettyIndent = ds.prettyIndent;
        ds.prettyPrinting = false;
        ds.prettyIndent = 0;
        XML.setSettings(ds);
    };

    this._restoreSettings = function() {
        if ('prettyPrinting' in this.XML_saved_settings && 'prettyIndent' in this.XML_saved_settings) {
            var ds = XML.settings();
            ds.prettyPrinting = this.XML_saved_settings.prettyPrinting;
            ds.prettyIndent = this.XML_saved_settings.prettyIndent;
            XML.setSettings(ds);
        }
        
        if (typeof(this.orig_max_script_run_time) == "number") {
            var _mozPref = new this.MozPref();
            _mozPref.SetIntPref('dom.max_script_run_time', this.orig_max_script_run_time);
            delete this.orig_max_script_run_time;
        }
    };

    this.pushHooksOff = function() {
        var console = this.console;
        var saved_hooks = [
                           console.jsds.topLevelHook,
                           console.jsds.functionHook,
                           console.jsds.breakpointHook,
                           console.jsds.debuggerHook,
                           // console.jsds.errorHook,
                           console.jsds.interruptHook,
                           //console.scriptHook.scriptCreatedName,
                           //console.scriptHook.scriptDestroyedName
                           ];
        this.saved_hook_stack.push(saved_hooks);
        console.jsds.topLevelHook = null;
        console.jsds.functionHook = null;
        console.jsds.breakpointHook = null;
        console.jsds.debuggerHook = null;
        // console.jsds.errorHook = null;
        this.clearInterruptHook();
        //console.scriptHook.scriptCreatedName = "jsdScriptCreated_hooksOff";
        //console.scriptHook.scriptDestroyedName = "jsdScriptDestroyed_hooksOff";
    };

    this.restoreHooks = function() {
        var console = this.console;
        var saved_hooks = this.saved_hook_stack.pop();
        var i = -1;
        console.jsds.topLevelHook = saved_hooks[++i];
        console.jsds.functionHook = saved_hooks[++i];
        console.jsds.breakpointHook = saved_hooks[++i];
        console.jsds.debuggerHook = saved_hooks[++i];
        // console.jsds.errorHook = saved_hooks[++i];
        this.setInterruptHook(saved_hooks[++i], "saved hook");
        // console.jsds.interruptHook = saved_hooks[++i];
        //console.scriptHook.scriptCreatedName = saved_hooks[++i];
        //console.scriptHook.scriptDestroyedName = saved_hooks[++i];
    };

    // Describe the event hooks here.
    this._script_event = function(evt, jsdScript) {


















    };

    this.jsdScriptCreated = function(jsdScript) {
        var console = this.console;
        try {
            var url = this.mainObj.utils.canonicalize_uri(jsdScript.fileName);
            if (url.indexOf("chrome://") == -1
                || !this.gave_message[url + "jsdScriptCreated"]) {
                this._script_event("jsdScriptCreated", jsdScript);



                this.gave_message[url + "jsdScriptCreated"] = true;
            }
            var sm = this.internURL(url);
            sm.onScriptCreated(jsdScript);
        } catch(ex) {
            this.log.exception(ex);
        }
    };

    this.internURL = function(url) {
        var sm;
        if (!(url in this.console.scriptManagers)) {
            sm = this.console.scriptManagers[url] = new JSDBGP.Scripts.ScriptManager(this.mainObj.scripts, url);
        } else {
            sm = this.console.scriptManagers[url];
        }
        return sm;
    };

    this.jsdScriptDestroyed = function(jsdScript) {
        var console = this.console;
        try {
            var url = this.mainObj.utils.canonicalize_uri(jsdScript.fileName);
            if (url.indexOf("chrome://") == -1
                || !this.gave_message[url + "jsdScriptDestroyed"]) {
                this._script_event("jsdScriptDestroyed", jsdScript);
                this.gave_message[url + "jsdScriptDestroyed"] = true;
            }
            var sm = console.scriptManagers[url];
            var tag = jsdScript.tag;
            if (!sm) {



                if (tag in this.pendingScripts) {



                    delete this.pendingScripts[tag];
                } else {



                }
                return;
            }
            sm.onScriptInvalidated(jsdScript);
        } catch(ex) {
            this.log.exception(ex);
        }
    };

    this.jsdScriptCreated_prologue = function(jsdScript) {
        this._script_event("jsdScriptCreated_prologue", jsdScript);
        this.pendingScripts[jsdScript.tag] = jsdScript;
        if (!this.setInitialInterrupt) {
            this.setInitialInterrupt = true;
            this.console.jsds.pause();
            this.setInterruptHook(this.console.executionHook_prologue,
                                  "executionHook_prologue");
            // this.console.jsds.interruptHook = this.console.executionHook_prologue;
            this.console.jsds.unPause();
        }
    };

    this.jsdScriptDestroyed_prologue = function(jsdScript) {
        this._script_event("jsdScriptDestroyed_prologue", jsdScript);
        delete this.pendingScripts[jsdScript.tag];
    };

    // We always have to keep processing scripts, but if we're in the
    // heart of the command-loop we'll just log them and walk the list
    // in order received outside the command-loop.
    
    this.jsdScriptCreated_hooksOff = function(jsdScript) {
        this._script_event("jsdScriptCreated_hooksOff", jsdScript);
        this.hooksOffScripts.push([true, jsdScript]);
    };

    this.jsdScriptDestroyed_hooksOff = function(jsdScript) {
        this._script_event("jsdScriptDestroyed_hooksOff", jsdScript);
        this.hooksOffScripts.push([false, jsdScript]);
    };

    this.processHooksOffScripts = function() {
        for (var i in this.hooksOffScripts) {
            var info = this.hooksOffScripts[i];
            if (info[0]) {



                var jsdScript = info[1];
                if (jsdScript.isValid) {
                    this.jsdScriptCreated(jsdScript);
                }
            } else {





                if (jsdScript.isValid) {
                    this.jsdScriptDestroyed(jsdScript);
                }
            }
        }
        this.hooksOffScripts = [];
    };

    // Hooks

    this.jsdExecutionHook_prologue = function(frame, type, rv) {
        var console = this.console;
        if (!console) {
            if (!this.jsdExecutionHook_prologue.gave_message) {



                this.jsdExecutionHook_prologue.gave_message = true;
                this.setStopState(false);
            }
            return JSDBGP.Constants.RETURN_CONTINUE;
        }



        if (!console.initialized) {



            Components.classes[JSDBGP.Constants.JSD_CTRID].getService(Components.interfaces.jsdIDebuggerService).interruptHook = null;
            return JSDBGP.Constants.RETURN_CONTINUE;
        }
        // Update everything at this point
        this.setInterruptHook(console.executionHook, "executionHook");

        //ASSERT(console.jsds.scriptHook == console.scriptHook);

        console.scriptHook.scriptCreatedName = "jsdScriptCreated";
        console.scriptHook.scriptDestroyedName = "jsdScriptDestroyed";
        this.processPendingScripts();
        // And go do the regular hook
        return this.jsdExecutionHook(frame, type, rv);    
    };

    this.processPendingScripts = function() {
        for (var tag in this.pendingScripts) {
            this.jsdScriptCreated(this.pendingScripts[tag]);
        }
        this.pendingScripts = {};
    };

    this.jsdBreakpointHook = function(frame, type, rv) {



        var str = "Hit jsdBreakpointHook, file ";
        if (!frame.script) {
            str += "unknown";
        } else {
            if (frame.script.fileName) {
                str += frame.script.fileName;
            } else {
                str += "????";
            }
            str += (", tag " + (frame.script.tag || "unknown"));
        }



        return this.jsdExecutionHook(frame, type, rv);
    };

    this.jsdExecutionHook = function(frame, type, rv) {
        var console = this.console;
        var hookReturn = JSDBGP.Constants.RETURN_CONTINUE;
        if (!console) {



            this.setStopState(false);
            return hookReturn;
        }
        var utils = this.mainObj.utils;
        try {
            if (!console.initialized) {




                this._clearHooks(Components.classes[JSDBGP.Constants.JSD_CTRID].getService(Components.interfaces.jsdIDebuggerService));
                this.detachDebugger();
                return hookReturn;
            }





            var fname;
            // Should we be here?
            if (!this.testAndLog(!("frames" in console),
                                  "Execution hook called while stopped") ||
                frame.isNative ||
                !this.testAndLog(frame.script,
                                  "Execution hook called with no script") ||
                !this.testAndLog(!(frame.script.flags & JSDBGP.Constants.SCRIPT_NODEBUG),
                                 "Stopped in a script marked as don't debug")) {
                return hookReturn;
            } else {
                var filter_reason = 0;
                if (utils.isURLDBGP(fname = frame.script.fileName)) {
                    filter_reason = 1;
                } else if (utils.isURLFiltered(fname)) {
                    filter_reason = 2;
                } else if (utils.cantGetSource(fname)) {
                    filter_reason = 3;
                }
                if (filter_reason > 0) {


















                    return hookReturn;
                }
            }
    
            // Build the array of frames by walking the call stack
            var frames = new Array();
            var prevFrame = frame;
            while (prevFrame) {
                frames.push(prevFrame);
                prevFrame = prevFrame.callingFrame;
            }
            console.frames = frames;
            // This next routine checks for breakpoints
            hookReturn = this.debugTrap(frames, type, rv);
        } catch(ex) {
            this.log.exception("executionHook failed to run: " + ex);
        }
        delete console.frames;
        return hookReturn;
    };
    
    // This is called only when we initiate a StepOver and StepOut
    // Not called on StepInto

    this.jsdCallHook = function(frame, type) {
        var console = this.console;
        try {
            if (!console.initialized) {
                var jsds = Components.classes[JSDBGP.Constants.JSD_CTRID].getService(Components.interfaces.jsdIDebuggerService);
                jsds.functionHook = null;
                jsds.topLevelHook = null;
                return;
            }
            // Don't count filtered URLs: native, chrome, and debugger
            if (!frame.script) {
                return;
            }
            var fname = frame.script.fileName;
            var utils = this.mainObj.utils;
            if (utils.isURLDBGP(fname) || utils.isURLFiltered(fname)
                || utils.cantGetSource(fname)) {







                return;
            }
            if (type == Components.interfaces.jsdICallHook.TYPE_FUNCTION_CALL) {
                if (console.check_for_run_end_only) {
                    return;
                }







        
                // This would be a good place to check the dbgp command session
                // once in a while to see if the user pressed break.

                // But this hook isn't called when user does 'Go'
        
                // If we're stepping in somewhere, turn off the breakpoint.
        
                this.setStopState(false);
        



            } else if (type == Components.interfaces.jsdICallHook.TYPE_FUNCTION_RETURN) {
                // we're called *before* the returning frame is popped from the
                // stack, so we want our depth calculation to be off by one.

                // Check to see if this was fired from a top-level
                // run command.
                if (console.check_for_run_end_only) {
                    try {
                        if (!frame.callingFrame) {
                            setTimeout(window.getAttention, 1);
                        }
                    } catch(ex) {
                    }
                    return;
                }
                
                var depth = -1;
                var prevFrame = frame;
                while (prevFrame) {
                    depth++;
                    prevFrame = prevFrame.callingFrame;
                }





        
                if (console._stepOverDepth && depth <= console._stepOverDepth) {



                    // It would be cool to reinstate the _stepPast
                    // check so we don't just bounce into a line
                    // we've already been at.
                    //
                    // This should handle nested functions, like
                    // f_main(f_arg1(), f_arg2()):
                    //
                    // step_out from f_arg1 should stop at the
                    // first line in f_arg2
                    
                    this.setStopState(true);
                    delete console._stepOverDepth;
                    // Set this variable so we don't stop at the line
                    // containing the function call.
                    var callingFrame = frame.callingFrame;
                    if (callingFrame && callingFrame.isValid) {
                        var _fileName;
                        try {
                            _fileName = callingFrame.script.fileName;
                        } catch(ex) {
                            _fileName = "?";
                        }
                        console._stepPast = {
                            line : this.getCurrentLineByMode(callingFrame),
                            frames_length : depth,
                            fileName : _fileName
                        }
                    }
                }
            }
            // Top-level hooks aren't very informative.
        } catch(ex) {
            this.log.exception(ex);
        }
    };

    //Todo: set stop state if we're doing something else
    this.jsdErrorHook = function(message, fileName, line, pos, flags, exception) {









        var console = this.console;
        try {
            if (!console.initialized || this.mainObj.utils.isURLFiltered(fileName)) {
                Components.classes[JSDBGP.Constants.JSD_CTRID].getService(Components.interfaces.jsdIDebuggerService).errorHook = null;
                return true;
            }
            this.log.error("Error at file "
                           + fileName
                           + ", line "
                           + line
                           + " -- " + message);
        } catch(ex) {
            this.log.exception(ex);
        }
        return true;
    };

/* The script classes are in winger-scripts.js */

// First cut: use javascript prompt function to get commands

    this.debugTrap = function(frames, type, rv) {
        var console = this.console;
        if (!console) {



            this.setStopState(false);
            return JSDBGP.Constants.RETURN_CONTINUE;
        }
        var jsdIExecutionHook = Components.interfaces.jsdIExecutionHook;
        var frame = frames[0];
        var retcode = JSDBGP.Constants.RETURN_CONTINUE;
        if (!frame.isValid) {



            return retcode;
        }
        
        try {













            switch (type) {
            case jsdIExecutionHook.TYPE_BREAKPOINT:
                var bpResult = this.testBreakpoint(frame, rv);
                if (bpResult != -1) {
                    return bpResult;
                }
                break;
            case jsdIExecutionHook.TYPE_DEBUG_REQUESTED:
                break;
            case jsdIExecutionHook.TYPE_DEBUGGER_KEYWORD:
                break;
            case jsdIExecutionHook.TYPE_THROW:
                return jsdIExecutionHook.RETURN_CONTINUE_THROW;
            case jsdIExecutionHook.TYPE_INTERRUPTED:
                // Step-into: step at the next line

                if (console._stepPast) {
                    // Did we move to a different line?
                    // This sequence is handled for step-into and break now
                    var fileName = frame.script.fileName;
                    var line = this.getCurrentLineByMode(frame);
                    if (console._stepPast.line == line
                        && console._stepPast.frames_length == frames.length
                        && console._stepPast.fileName == fileName) {














                        return retcode;
                    } else {
                        delete console._stepPast;
                    }
                }

                if (!this.sentInitString_p) {
                    console.command.sendInitString(JSDBGP.Constants.STOP_REASON_BREAK);
                    this.sentInitString_p = true;
                }

                break;
            default:
                /* don't print stop/cont messages for other types */
            }
            console.command.wrapDoStatus("break");
            if (console.command.breakClosure) {
                console.command.breakClosure();
                console.command.breakClosure = null;
            }
            this.setStopState(false);
            
            try {





                console.command.doEnterNestedEventLoop();
            } catch(ex) {
                this.log.exception(ex, "Problem doing enter-nested-event");
            }
        } catch(ex) {
            this.log.exception(ex);
        }
        return retcode;
    };

    this.clearStepPast = function() {
        delete this.console._stepPast;
        delete this.console._stepOverDepth;
    }

    this.getCurrentLineByMode = function(frame) {
        if (!this.mainObj.settings.settings['pretty_print'][0]) {
            return frame.line;
        }
        try {
            var line = frame.script.pcToLine(frame.pc, frame.script.PCMAP_PRETTYPRINT);



            return line;
        } catch(ex) {



            return frame.line;
        }
    };

    this.commence_shutdown = function() {
         // Wait until we return out of the program to shut down.
        function func(self) {
            return (function() {
                return self.detachDebugger();
            });
        }
        setTimeout(func(this), 200);
    };

    this.testBreakpoint = function(currentFrame, rv) {
        var jsdScript = currentFrame.script;
        if (!jsdScript.isValid) {
            return JSDBGP.Constants.RETURN_CONTINUE;
        }
        var tag = jsdScript.tag;
        return this.mainObj.breakpointManager.testBreakpoint(tag, currentFrame, -1);
    };

    this._getScriptsByTag = function() {
        var scriptsByTag = {};
        var scriptManagers = this.console.scriptManagers;
        for (var url in scriptManagers) {
            var sm = scriptManagers[url];
            for (var tag in sm.scriptWrappers) {
                var jsdScript = sm.scriptWrappers[tag].jsdScript;
                if (jsdScript.isValid) {
                    scriptsByTag[tag] = sm.scriptWrappers[tag].jsdScript;
                }
            }
        }
        return scriptsByTag;
    };

    this._updateNewScripts = function() {
        var currScripts = this._getScriptsByTag();









        // First delete all invalid existing scripts that aren't in the new list
        for (var tag in currScripts) {
            var jsdScript = currScripts[tag];
            if (!jsdScript.isValid) {



                this.jsdScriptDestroyed(jsdScript);
                delete currScripts[tag];
            };
        }
        // Now add all scripts we just learned about






        for (var tag in this.newScriptsByTag) {
            if (!(tag in currScripts)) {







                this.jsdScriptCreated(this.newScriptsByTag[tag]);
            }
        }
        this.newScriptsByTag = {};



    };

    this.eventLoopNested = function() {



        // do a stop
        // console.command.dispatch('break');
    };

    this.setStopState = function(state) {
        var console = this.console;



        if (state) {
            this.setInterruptHook(console.executionHook, "standard");
        } else {
            this.clearInterruptHook();
        }
    };

    this.testAndLog = function(pred, comment) {
        if (!pred) {



        }
        return pred;
    };

    this.clearInterruptHook = function() {



        this.console.jsds.interruptHook = null;
    }

    this.setInterruptHook = function(func, label) {












        this.console.jsds.interruptHook = func;
    }
};
